#!/usr/bin/python2 

import socket
import struct
import sys
import json
import os
import pickle
import rsa
import time
import hashlib



def main():
        f = open("network_settings.json", "r") #get network settings from the file and read them into memory
        settings = json.load(f)
        f.close()


        maxattempts = 5 #make 5 attempts to connect to the client
        ip_address = settings['Server IP']
        port = settings['Authentication Port']        
        integer_size = struct.calcsize("i")
        comm_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM) #open a TCP socket
        comm_sock.setblocking(True)
        attempts = 0;

        
        while(True):
                try:
                        print("Connecting to host at: "+ip_address+":"+str(port)) #connect to the client process at the specified IP and port
                        comm_sock.connect((ip_address, port)) #attempt to connect to the server, try 5 times and quit program if all attempts fail
                        print("Connection successful")
                        break
                except:
                        if(attempts == maxattempts):
                                print("Connection timed out.")
                                return 0
                        attempts += 1
                        print("Connection to the host failed.")
                        print("Attempting to reconnect. Attempt "+str(attempts)+" out of "+str(maxattempts))
                        time.sleep(5)

        aes_key = os.urandom(16) #generate AES-256 key using built-in OS RNG
        hmac_key = os.urandom(16) #do the same for the HMAC key

        (pubkey, privkey) = rsa.newkeys(2048, poolsize=8) #generate an RSA key pair
        privkey_s = pickle.dumps(privkey) #serialize both the private and public key
        pubkey_s = pickle.dumps(pubkey)
        
        
        length_pubkey = len(pubkey_s)
        
        comm_sock.sendall(struct.pack("i", length_pubkey) + pubkey_s) #send public key to server
        
        
        data = b""
        
        while len(data) < integer_size: #wait for the server's public key
                data += comm_sock.recv(integer_size) 
        key_size = data[:integer_size]
        key_size = struct.unpack("i", key_size)[0]
        data = data[integer_size:]
        
        while len(data) < key_size: 
                data += comm_sock.recv(key_size)
                
        key_s = data[:key_size]
        key = pickle.loads(key_s) #public key from server
        data = data[key_size:]
        
        print("Enter password") 
        hash1 = hashlib.sha256() #enter password and immediately hash it
        password = input().encode('utf-8')
        hash1.update(password)
        password = hash1.digest()
        password_e = rsa.encrypt(password, key) #encrypt the password with the private key
        password_length = len(password_e)
        #print("bytes sent:")
        #print(password_e)
        comm_sock.sendall(struct.pack("i", password_length) + password_e) #send the encrypted password over TCP
        
        while len(data) < integer_size: #wait for a response
                data += comm_sock.recv(integer_size) 
        pass_confirm = data[:integer_size]
        pass_confirm = struct.unpack("i", pass_confirm)[0]
        #print(pass_confirm)
        data = data[integer_size:]
        
        if pass_confirm == 99: #if the password is accepted
                encrypted_aes_key = rsa.encrypt(aes_key, key) #encrypt the AES key
                length_aes_key = len(encrypted_aes_key)

                comm_sock.sendall(struct.pack("i", length_aes_key) + encrypted_aes_key) #send the encrypted AES key
                
                encrypted_hmac_key = rsa.encrypt(hmac_key, key) #encrypt the HMAC key
                length_hmac_key = len(encrypted_hmac_key)

                comm_sock.sendall(struct.pack("i", length_hmac_key) + encrypted_hmac_key) #send the encrypted HMAC key 
                
                
                aes_key_file = open("aes_key.txt", "wb") #write the AES and HMAC keys to a file
                aes_key_file.write(aes_key)
                aes_key_file.close()
             
                hmac_key_file = open("hmac_key.txt", "wb")
                hmac_key_file.write(hmac_key)
                hmac_key_file.close()
                return 1

        else:
                print("Password not accepted! Try again!") #password was not correct, tell user
                return 0
                        
if __name__ == "__main__":
        main()

