#!/usr/bin/python2 

import socket
import struct
import sys
import json
import os
import pickle
import rsa
import time
import hashlib
import hmac
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad



def main():
        f = open("network_settings.json", "r") #load the network settings into memory
        settings = json.load(f)
        f.close()

        maxattempts = 5
        ip_address = settings['Server IP']
        port = settings['Configuration Port']
        
        float_size = struct.calcsize("f")
        integer_size = struct.calcsize("i")
        
        comm_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM) #create a TCP socket
        comm_sock.setblocking(True) 
        try:
                aes_key_file = open("aes_key.txt", "rb") #load AES and HMAC keys into memory
                aes_key = aes_key_file.read()
                aes_key_file.close()
                print(aes_key)
                
                hmac_key_file = open("hmac_key.txt", "rb")
                hmac_key = hmac_key_file.read()
                hmac_key_file.close()
        except:
                print("Encryption keys missing!") #quit program if encryption keys are missing
                return
        
        attempts = 0;
        
        
        while(True):
                try:
                        print("Connecting to host at: "+ip_address+":"+str(port)) #connect to server host at the specified IP and port
                        comm_sock.connect((ip_address, port))
                        print("Connection successful")
                        break
                except:
                        if(attempts == maxattempts):
                                print("Connection timed out.")
                                return
                        attempts += 1
                        print("Connection to the host failed.")
                        print("Attempting to reconnect. Attempt "+str(attempts)+" out of "+str(maxattempts))
                        time.sleep(5)
        try:


                
                
                while 1:
                        data = b"" #user is given options
                        print("Please select a camera setting to update:")
                        print("1) Camera resolution")
                        print("2) Motion-triggered video length")
                        print("3) Live streaming")
                        print("4) Image saturation")
                        print("5) Image contrast")
                        print("6) Encryption")
                        print("7) Motion sensitivity")
                        print("8) Change authentication password")
                        print("9) Exit program")

                        try:
                                option = int(input())
                        except KeyboardInterrupt:
                                break
                        except:
                                option = -1
                        
                        if option == 1: #user has elected to change resolution
                                comm_sock.sendall(struct.pack("i", option)) #the option is sent to the server over TCP
                        
                                while len(data) < integer_size: #wait for the current resolution settings to be returned to the client
                                        data += comm_sock.recv(integer_size) 
                                current_setting1 = data[:integer_size]
                                current_setting1 = struct.unpack("i", current_setting1)[0]
                                data = data[integer_size:]                      
                                
                                while len(data) < integer_size:
                                        data += comm_sock.recv(integer_size) 
                                current_setting2 = data[:integer_size]
                                current_setting2 = struct.unpack("i", current_setting2)[0]
                                data = data[integer_size:]
                                

                                print("Current setting:", str(current_setting1)+"x"+str(current_setting2)) #display the current resolution
                                print("Enter x")
                                new_setting1 = int(input()) #user will enter x setting
                                print("Enter y")
                                new_setting2 = int(input()) #then the y setting
                                mac = hmac.new(hmac_key, str(new_setting1+new_setting2).encode('utf-8'), hashlib.sha256) 
                                print(len(mac.hexdigest()))
                                comm_sock.sendall(struct.pack("i", new_setting1)+struct.pack("i", new_setting2)+mac.hexdigest().encode('utf-8')) #send the new settings and their associated HMAC
                                while len(data) < integer_size: #wait for a confirmation
                                        data += comm_sock.recv(integer_size) 
                                confirmation_code = data[:integer_size]
                                confirmation_code = struct.unpack("i", confirmation_code)[0]
                                data = data[integer_size:]
                                        
#                                print("Confirm code", str(confirmation_code))
                                        
                                if (confirmation_code == 100): #if the change was successful
                                        print("Setting changed successfully") #continue program loop
                                elif (confirmation_code == 77): #authentication (HMAC) failure
                                        print("Unauthentic client detected!") 
                                        return #end program
                                else: #if the change was not successful
                                        print("Configuration failed.") #continue program loop

                        elif (option >= 0) and (option < 8): #all other single-input options
                                comm_sock.sendall(struct.pack("i", option)) #tell the server which option is being changed

                                while len(data) < float_size: #wait for server to send current settings
                                        data += comm_sock.recv(float_size) 
                                current_setting = data[:float_size]
                                current_setting = struct.unpack("f", current_setting)[0]
                                data = data[float_size:]                        
                                print("Current setting:", str(current_setting)) #print for current settings
                                new_setting = float(input()) #user inputs new setting
                                mac = hmac.new(hmac_key, str(new_setting).encode('utf-8'), hashlib.sha256)                              
                                comm_sock.sendall(struct.pack("f", new_setting) + mac.hexdigest().encode('utf-8')) #send new setting and its associated HMAC
                                while len(data) < integer_size: #wait for a confirmation code
                                        data += comm_sock.recv(integer_size) 
                                confirmation_code = data[:integer_size]
                                confirmation_code = struct.unpack("i", confirmation_code)[0]
                                data = data[integer_size:]      
                                
                                #print("Confirm code", str(confirmation_code))
                                #print("New setting", str(new_setting))
                
                                if (confirmation_code == 100): #if change was successful
                                        print("Setting changed successfully")
                                elif (confirmation_code == 77): #if change was not successful
                                        print("Unauthentic client detected!")
                                        return
                                else:
                                        print("Configuration failed.")  
                                        
                        elif (option == 8): #if user wants to change the password
                                comm_sock.sendall(struct.pack("i", option)) #tell the server that the password is to be changed
                                print("Enter previous password")
                                old_password = input().encode('utf-8') #enter the current password and hash it
                                hash1 = hashlib.sha256()
                                hash1.update(old_password)
                                old_password = hash1.digest()
                                iv = os.urandom(16)
                                AES_cipher = AES.new(aes_key, AES.MODE_CBC, iv) 
                                #print(AES_cipher)
                                old_password_e = AES_cipher.encrypt(pad(old_password,16)) #encrypt it with AES-256
                                old_password_size = len(old_password_e)
                                mac = hmac.new(hmac_key, old_password_e, hashlib.sha256).hexdigest() #attach an HMAC
                                #print(aes_key)
                                comm_sock.sendall(struct.pack("i", old_password_size) + old_password_e + iv + mac.encode('utf-8')) #send it to the server
                                
                                while len(data) < integer_size: #wait for a confirmation
                                        data += comm_sock.recv(integer_size) 
                                password_confirmed = data[:integer_size]
                                password_confirmed = struct.unpack("i", password_confirmed)[0]
                                data = data[integer_size:]      

                                if password_confirmed == 678: #if the password is accepted
                                        while 1:
                                                print("Enter new password") #have the user enter a new password
                                                new_password = input().encode('utf-8')
                                                print("Enter new password again to confirm")    #and confirm it
                                                new_password_check = input().encode()
                                                if new_password_check == new_password: #if confirmed, break loop
                                                        break
                                                else:
                                                        print("Passwords do not match! Try again!") #user must enter two identical passwords

                                        if new_password_check == new_password:
                                                hash1 = hashlib.sha256() #hash password, encrypt it, add an HMAC, and send it over TCP
                                                hash1.update(new_password)
                                                new_password = hash1.digest()
                                                iv = os.urandom(16)
                                                AES_cipher = AES.new(aes_key, AES.MODE_CBC, iv)
                                                new_password = AES_cipher.encrypt(pad(new_password,16))
                                                new_password_size = len(new_password)
                                                mac = hmac.new(hmac_key, new_password, hashlib.sha256).hexdigest()
                                                comm_sock.sendall(struct.pack("i", new_password_size) + new_password + iv + mac.encode('utf-8'))
                                        
                                                while len(data) < integer_size: #wait for a response
                                                        data += comm_sock.recv(integer_size) 
                                                success = data[:integer_size]
                                                success = struct.unpack("i", success)[0]
                                                data = data[integer_size:]      
                                                if success == 982: #password changed successfully
                                                        print("Password changed successfully!")
                                                else: #error occured
                                                        print("Error. Password not changed.")

                                elif password_confirmed == -1: #HMAC authentication error
                                        print("Unauthentic client detected!")
                                        return
                                else: #the current password entry was not correct
                                        print("Password incorrect! Please try again!")          
                                
                        
                        elif (option == 9): #end program
                                print("Exiting program...")
                                return
                        
                        else: #option does not exist
                                print("Invalid option selected. Please try again.")
                                
                        
                        
                

                
                
        except Exception as e:
                print(e)
                print("Invalid command")



                        
if __name__ == "__main__":
        main()

