import socket
import time
import sys
import cv2
import pickle
import numpy as np
import struct
import math
import json
import hmac
import hashlib
import rsa
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad


def getTime():
        now = time.time()
        return now



def main():             
        try:
                f = open("network_settings.json") #load network settings into memory
                settings = json.load(f)
                host = settings["Server IP"]
                port = settings["Stream Port"]
        except Exception as e:
                print(e)
                print("Settings file corrupt or missing!")
                return

        aes_key_file = open("aes_key.txt", "rb") #load AES and HMAC keys into memory
        aes_key = aes_key_file.read()
        aes_key_file.close()

        hmac_key_file = open("hmac_key.txt", "rb")
        hmac_key = hmac_key_file.read()
        hmac_key_file.close()
        

        sock=socket.socket(socket.AF_INET,socket.SOCK_DGRAM) #create UDP socket
        sock.setblocking(True)
        print('Socket created')

        sock.bind(("",port))

        print("Press Control + C to exit the program.")

        integer_size = struct.calcsize("i")
        float_size = struct.calcsize("d")

        constructed_image = b""
        received_seqn = []
        received_seqn_ordered = []
        last_seqn = -1
        buffer = 65536 #we want to read 65536 bits at a time
        max_size = 5
        sequence_counter = 0
        mask = [1,2,3,4,5]
        while True:
        	#start receiving packets ovver UDP
                data = b"" #start with blank memory
                rightnow = getTime()
                while len(data) < 1: #check for encryption bit
                        receivedData = sock.recvfrom(buffer)
                        data += receivedData[0]
                encryption_status = int(data[:1])
                #print("Encryption:",str(encryption_status))
                data = data[1:]
                while len(data) < integer_size:
                        receivedData = sock.recvfrom(buffer)
                        data += receivedData[0]
                packed_img_size = data[:integer_size] #get size of message
                data = data[integer_size:]
                img_size = struct.unpack("i", packed_img_size)[0]
                rightnow = getTime()
                while len(data) < float_size: #get timestamp
                        receivedData = sock.recvfrom(buffer)
                        data += receivedData[0]
                time_data = data[:float_size]
                data = data[float_size:]
                timestamp = struct.unpack("d", time_data)[0]
                #print(timestamp)
                while len(data) < integer_size: #get the packet number
                        receivedData = sock.recvfrom(buffer)
                        data += receivedData[0]
                seqn = data[:integer_size]
                seqn = struct.unpack("i", seqn)[0]
                #print(seqn)
                data = data[integer_size:]

                while len(data) < integer_size: #get max number of packets in sequence
                        receivedData = sock.recvfrom(buffer)
                        data += receivedData[0]
                num_of_segs = data[:integer_size]
                num_of_segs = struct.unpack("i", num_of_segs)[0]
                data = data[integer_size:]
                #print(num_of_segs)
                if (num_of_segs != max_size): #if the received packet's number is less than the max packets in the sequence, keep track of it
                        max_size = num_of_segs 
                        mask = []
                        for k in range(1,max_size+1):
                                mask.append(k)
                        received_seqn_ordered = []
                        sequence_counter = 0
                        
                while len(data) < 16: #get AES IV
                        receivedData = sock.recvfrom(buffer)
                        data += receivedData[0]
                iv = data[:16]
                
                data = data[16:]

                while len(data) < img_size: #receive the image data
                        receivedData = sock.recvfrom(buffer)       
                        data += receivedData[0]
                if img_size > 0: #do nothing if the received data is at least 1 bit
                        pass
                else: #if not, reset
                        constructed_image = b"" 
                        received_seqn_ordered = []
                        sequence_counter = 0
                        
                #delay = abs(getTime()-timestamp)*1000
                #print("Delay "+str(delay)+" ms")
                        
                data = data[:img_size]


                if seqn == sequence_counter + 1: #add image data of each packet together if they are of consecutive sequence numbers
                        #print(seqn)
                        constructed_image = constructed_image + data
                        received_seqn_ordered.append(seqn)
                        sequence_counter += 1                                
                
                #print(seqn)
                #print(received_seqn_ordered)
                #print(aes_key)
                
                if mask == received_seqn_ordered: #if the sequence is complete
                        try:
                                #print("current IV", str(iv))
                                if encryption_status: #If encryption is enabled
                                        aes_cipher = AES.new(aes_key, AES.MODE_CBC, iv)
                                        constructed_image = unpad(aes_cipher.decrypt(constructed_image), 16) #decrypt image
                                frame = pickle.loads(constructed_image[64:]) #load image
                                mac = constructed_image[:64].decode('utf-8')
                                if mac == hmac.new(hmac_key, constructed_image[64:], hashlib.sha256).hexdigest(): #check HMAC
                                        frame = cv2.imdecode(frame, cv2.IMREAD_COLOR) #decode frame
                                else:
                                        print("Image could not be authenticated!") #HMAC failed
                                received_seqn_ordered = [] #reset sequence counter
                                constructed_image = b""
                                #sock.sendto(bytes(str(getTime()), 'utf-8'), ("192.168.0.87", port)) #send time to host, unused
                                cv2.imshow('Feed from camera',frame) #show the received image in frame
                                if cv2.waitKey(1) & 0xFF == ord('x'):
                                        break
                                sequence_counter = 0
                        except Exception as e: #reset sequence counter
                                #print("Error handled")
                                #print(str(e))
                                constructed_image = b""
                                sequence_counter = 0
                                received_seqn_ordered = []
                elif len(received_seqn_ordered) == max_size: #reset the sequence counter once sequence is complete
                        received_seqn_ordered = []
                        constructed_image = b""
                        sequence_counter = 0

                last_seqn = seqn #Keep track of last sequence number
                #print("mask")
                #print(mask)
                                
                        
if __name__ == "__main__":
        main()
