#!/usr/bin/python2 

import socket
import struct
import sys
import json
import os
import pickle
import rsa
import time
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad
import hmac
import hashlib

def main():
	f = open("network_settings.json", "r") #open network settings file and read its contents
	settings = json.load(f)
	f.close()

	maxattempts = 5 #set maximum number of connection attempts
	ip_address = settings['Server IP'] #set IP from network settins file
	port = settings['Video Port'] #set port from same file
 
	aes_key_file = open("aes_key.txt", "rb") #retrieve AES symmetric key for decryption of video data
	aes_key = aes_key_file.read()
	aes_key_file.close()
	
	hmac_key_file = open("hmac_key.txt", "rb") #retrieve HMAC key for data integrity verification
	hmac_key = hmac_key_file.read()
	hmac_key_file.close()

	integer_size = struct.calcsize("i") #set integer size for struct
	comm_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM) #open TCP socket 
	comm_sock.setblocking(True) #set a blocking socket
	attempts = 0; #count connection attempts
	while(True): #attempt to connect five times, then quit if unsuccessful all five times
		try:
			print("Connecting to host at: "+ip_address+":"+str(port)) #connect to the server host at the given IP/port
			comm_sock.connect((ip_address, port))
			print("Connection successful")
			break
		except:
			if(attempts == maxattempts): #quit once too many attempts have been made
				print("Connection timed out.")
				return
			attempts += 1 #if fail, increment attempts counter
			print("Connection to the host failed.")
			print("Attempting to reconnect. Attempt "+str(attempts)+" out of "+str(maxattempts))
			time.sleep(5)
	try:


		
		
		while 1:
                        #give the user their options
			data = b""
			print("Please select an option below:")
			print("1) View video directory")
			print("2) Download video")
			print("3) Delete video")
			print("4) Exit program")
			option = int(input())

			#if the user wants to view the video directory
			if option == 1: #send a command to the server asking for the video list
				mac = hmac.new(hmac_key, str(16+77).encode('utf-8'), hashlib.sha256) #generate HMAC of command
				comm_sock.sendall(struct.pack("i", 16) + struct.pack("i", 77) + mac.hexdigest().encode('utf-8')) #send everything over TCP

				#wait to receive the size of the list over TCP
				while len(data) < integer_size:
					data += comm_sock.recv(integer_size) 
				list_size = data[:integer_size]
				list_size = struct.unpack("i", list_size)[0]
				data = data[integer_size:]
                                        
                                #if the size is -1, then the client has been deemed unauthentic due to failed HMAC verification
				if list_size == -1:
                                        print("Unauthentic client detected!")
                                        break
                                #receive AES initialization vector
				while len(data) < 16:
					data += comm_sock.recv(16) 
				iv = data[:16]
				data = data[16:]
				#print(iv)

				#receive the HMAC code for verification
				while len(data) < 64:
					data += comm_sock.recv(64)
				mac = data[:64].decode('utf-8')
				data = data[64:]

				#receive the video list
				while len(data) < list_size:
					data += comm_sock.recv(list_size)
				video_list_s = data[:list_size]
				print("received mac",str(mac))
				print("generated mac", str(hmac.new(hmac_key, video_list_s, hashlib.sha256).hexdigest())) #verify the data's integrity with HMAC
				if mac == hmac.new(hmac_key, video_list_s, hashlib.sha256).hexdigest():
					aes_cipher = AES.new(aes_key, AES.MODE_CBC, iv)
					video_list_s = unpad(aes_cipher.decrypt(video_list_s), 16) #decrypt the video data
					video_list = pickle.loads(video_list_s) #de-serialize the data back into a Python object
					if video_list['videos'] > 0: #if list is not empty
						for i in video_list: #print contents of list in a formatted way
							if i == "videos":
								print("Number of videos saved:",str(video_list[i]))
							else:
								print("Index:",str(i),"Filename:",str(video_list[i][0])+".mp4","Size:",str(video_list[i][1]),"Date created:",str(video_list[i][2]))
					else:
						print("No videos saved yet.") #if empty, tell the user
				else:   #if data is unauthentic, do nothing and tell the user
					print("Authentication failed")
					
			elif option == 4: #quit the program upon request by the user
				print("Exiting program...")
				break
			elif option == 2: #download a video from the server
				print("Enter index number of video to download:")
				option2 = int(input()) #user inputs the index of the video they wish to download from the video directory
				mac = hmac.new(hmac_key, str(45+option2).encode('utf-8'), hashlib.sha256) #generate HMAC
				comm_sock.sendall(struct.pack("i", 45) + struct.pack("i", option2) + mac.hexdigest().encode('utf-8')) #send this index plus the HMAC

				while len(data) < integer_size:
					data += comm_sock.recv(integer_size) 
				status_code1 = data[:integer_size]
				status_code1 = struct.unpack("i", status_code1)[0]
				data = data[integer_size:]

				if status_code1 == -1: #HMAC authentication error
                                        print("Unauthentic client detected!")
                                        return

				if (status_code1 == 450): #the requested video exists
					print("Video exists")
					data = b""
					while len(data) < integer_size: #get size of the video
						data += comm_sock.recv(integer_size) 
					video_size = data[:integer_size]
					video_size = struct.unpack("i", video_size)[0]
					data = data[integer_size:]
					while len(data) < 16: #get AES IV
						data += comm_sock.recv(16) 
					iv = data[:16]
					data = data[16:]
					#print(iv)
					print(3)
					while len(data) < 64: #get HMAC
						data += comm_sock.recv(64)
					mac = data[:64].decode('utf-8')
					data = data[64:]

					while len(data) < video_size: #get encrypted video
						data += comm_sock.recv(video_size)
					video_s = data[:video_size]
					print("received mac",str(mac))
					print("generated mac", str(hmac.new(hmac_key, video_s, hashlib.sha256).hexdigest()))
					if mac == hmac.new(hmac_key, video_s, hashlib.sha256).hexdigest(): #if HMAC is correct
						aes_cipher = AES.new(aes_key, AES.MODE_CBC, iv)
						video_s = unpad(aes_cipher.decrypt(video_s), 16) #decrypt the video
						print("Please enter a filename for the saved video") #the user will input a filename
						new_name = input()
						try:
							f = open(new_name+".mp4", "wb") #this video will be saved under the selected filename
							f.write(video_s)
							f.close()
							print(new_name, "saved successfully!")
						except:
							print("Write operation failed") #if the video was not successfully written
							
					else:
						print("Unauthentic video detected!") #the video's HMAC is incorrect
				else:
					print("Video at the index specified does not exist!")
                                        
				
			elif option == 3: #the user has requested to delete a video
				print("Enter index number of video to delete:")
				option2 = int(input()) #the user will enter the index of the video they wish to delete
				mac = hmac.new(hmac_key, str(78+option2).encode('utf-8'), hashlib.sha256)
				comm_sock.sendall(struct.pack("i", 78) + struct.pack("i", option2) + mac.hexdigest().encode('utf-8')) #send the index + an HMAC
				
				while len(data) < integer_size: #get a confirmation code
					data += comm_sock.recv(integer_size) 
				status_code1 = data[:integer_size]
				status_code1 = struct.unpack("i", status_code1)[0]
				data = data[integer_size:]

				if status_code1 == -1: #HMAC authentication failure
                                        print("Unauthentic client detected!")
                                        return
			
				if (status_code1 == 450): #video exists
					while len(data) < integer_size:
						data += comm_sock.recv(integer_size) 
					status_code2 = data[:integer_size]
					status_code2 = struct.unpack("i", status_code2)[0]
					data = data[integer_size:]
					
					if (status_code2 == 900): #video has been successfully deleted
						print("Video at index", str(option2), "deleted successfully")
					else: #video has not been deleted
						print("Video at index", str(option2), "was not deleted or does not exist.")
					
				else: #video does not exist
					print("Video at the index specified does not exist!")
			else:
				#HMAC authentication failure
				print("Invalid option selected. Please try again.")
			
		

		
		
	except Exception as e:
		print(e)



			
if __name__ == "__main__":
	main()

